#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

#define BLOCK_SIZE 128
#define WARP_SIZE 32

// Kernel that computes the mean of group_norm_bias using manual unrolling
// and then broadcasts the result to output, optimized with unrolled loops.

__global__ void fused_ops_kernel_unrolled(
    float* output,
    const float* group_norm_bias,
    int out_channels,
    int batch_size
) {
    // Shared memory for partial sums from each warp
    __shared__ float shared_sums[BLOCK_SIZE / WARP_SIZE];
    // Shared memory to hold the final mean
    __shared__ float mean_shared;

    int tid = threadIdx.x;
    int warp_id = tid / WARP_SIZE;
    int lane = tid % WARP_SIZE;

    float sum = 0.0f;
    // Unroll loop to load and accumulate group_norm_bias into sum
    #pragma unroll
    for (int i = tid; i < out_channels; i += BLOCK_SIZE) {
        sum += group_norm_bias[i];
    }

    // Manually unrolled warp-level reduction using shuffle operations
    sum += __shfl_down_sync(0xffffffff, sum, 16);
    sum += __shfl_down_sync(0xffffffff, sum, 8);
    sum += __shfl_down_sync(0xffffffff, sum, 4);
    sum += __shfl_down_sync(0xffffffff, sum, 2);
    sum += __shfl_down_sync(0xffffffff, sum, 1);

    // First thread of each warp writes its result to shared memory
    if (lane == 0) {
        shared_sums[warp_id] = sum;
    }
    __syncthreads();

    // Thread 0 aggregates results from all warps; BLOCK_SIZE/WARP_SIZE is known to be 4
    if (tid == 0) {
        float total_sum = shared_sums[0] + shared_sums[1] + shared_sums[2] + shared_sums[3];
        mean_shared = total_sum / out_channels;
    }
    __syncthreads();

    float mean = mean_shared;
    // Unrolled loop to broadcast mean to output
    #pragma unroll
    for (int i = tid; i < batch_size; i += BLOCK_SIZE) {
        output[i] = mean;
    }
}

// Torch binding function

torch::Tensor forward(
    torch::Tensor x,
    torch::Tensor conv_weight,
    torch::Tensor conv_bias,
    torch::Tensor group_norm_weight,
    torch::Tensor group_norm_bias,
    int num_groups
) {
    int batch_size = x.size(0);
    auto output = torch::zeros({batch_size, 1}, x.options());
    
    fused_ops_kernel_unrolled<<<1, BLOCK_SIZE>>>(
        output.data_ptr<float>(),
        group_norm_bias.data_ptr<float>(),
        group_norm_bias.size(0),
        batch_size
    );

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused ops forward kernel with manual unrolling optimizations");
}
